"""Tools for saving articles to a database"""

import os
from datetime import datetime, timedelta
from typing import Optional

import pandas as pd
from pymongo import DESCENDING, MongoClient
from pymongo.errors import ServerSelectionTimeoutError

from monito_tools import general_config, log, profile

_DB = None

columns = [
    "outlet",
    "url",
    "title",
    "subtitle",
    "text",
    "published",
    "published_iso",
    "saved",
    "saved_iso",
    "authors",
    "bag_of_words",
    "category",
]
articles_cache = None
last_update_articles_cache = None


@profile
def get_articles_cache(n_articles: int = 20000, dt: Optional[timedelta] = timedelta(seconds=60)) -> pd.DataFrame:
    """fill the global variable articles_cache with the last n articles"""

    global articles_cache, last_update_articles_cache
    if articles_cache is not None and dt is not None and last_update_articles_cache > datetime.now() - dt:
        return articles_cache

    articles_db = init_mongodb()

    if articles_cache is None:
        print("hier?")
        cursor = articles_db.find().sort("published", DESCENDING).limit(n_articles)
        print("dazwischen")
        df = pd.DataFrame(list(cursor), columns=columns)[columns]
        print("hä?")
        articles_cache = df
        print("echt?")
    else:
        print("before")
        cursor = articles_db.find({"saved_iso": {"$gt": last_update_articles_cache}})
        df = pd.DataFrame(list(cursor), columns=columns)[columns]
        articles_cache = pd.concat([df, articles_cache]).sort_values("published")
        print("after")

    last_update_articles_cache = datetime.now()

    first_date = df.loc[len(df) - 1, "published"]
    last_date = df.loc[0, "published"]
    log.debug(f"cached articles from {first_date} to {last_date}")

    print(df.loc[:5][["title", "published", "authors"]])
    print(df.loc[len(df) - 5 :][["title", "published", "authors"]])
    return df


def init_mongodb():
    """initialise the mongodb"""
    global _DB

    # log.debug("Getting database handle")
    if _DB is not None:
        return _DB

    # patch_all()

    if "MONITO_DB_URI" in os.environ:
        db_uri = os.environ["MONITO_DB_URI"]
        log.info("Found environment variable MONITO_DB_URI")
    else:
        db_uri = general_config["db"]
        log.info("Using db from general config")

    try:
        client = MongoClient(db_uri)
        database = client["monite"]
        articles_db = database["articles"]
    except ServerSelectionTimeoutError as e:
        log.error(f"ServerSelectionTimeoutError:\n{str(e)}")

    _DB = articles_db
    return articles_db


@profile
def save_article_to_mongodb(article_dict, config, articles_db, i: int = 0, n: int = 0, overwrite: bool = False) -> bool:
    """save an article to mongodb"""

    if art := articles_db.find_one(
        {
            "$and": [
                {"published": article_dict["published"]},
                {"outlet": article_dict["outlet"]},
                {"title": article_dict["title"]},
            ]
        }
    ):
        if overwrite:
            log.warning("The article already exists in the data base. Overwriting.\n\t" f"{art['url']}")
            articles_db.update_one({"_id": art["_id"]}, {"$set": article_dict})
            return True
        log.info("The article already exists in the data base. Not overwriting.\n\t" f"{art['url']}")
        return False

    try:
        articles_db.insert_one(article_dict)
        log.info(f"published: {article_dict['published']}")
        return True
    except Exception as e:
        log.error("Exception occurred during db insert: ")
        log.error(str(e))
        return False


# def get_urls_in_db(config: dict, articles_db) -> list:
#    """get urls that have already been processed"""
#    return articles_db.distinct("url")

# urls_in_db = []
# with open_fs(config["output_directory"]) as articles_fs:
#     urls_in_db = [
#         json.load(articles_fs.open(path))["url"]
#         for path in articles_fs.walk.files(filter=["*.json"])
#     ]
# return urls_in_db



"""Tools for accessing text metrics, sanitize text etc.
"""

import os
import re
import unicodedata
from datetime import datetime, timedelta
from glob import glob
from typing import List
from urllib.parse import quote, quote_plus, urlsplit, urlunsplit
from zoneinfo import ZoneInfo

import numpy as np
import spacy
from spacy_syllables import SpacySyllables  # noqa: ignore

from monito_tools import general_config, log


def preprocess(nlp, text: str) -> List[str]:
    """
    preprocess text using spacy
    """

    text: spacy.tokens.doc.Doc = nlp(text)
    bow, _, _, _ = get_bag_of_words(text)
    bow = [replace_accented_letters(tok) for tok in bow]

    return bow


def med(str1, str2):
    """minimum edit distance."""
    n = len(str1)
    m = len(str2)

    matrix = [[i + j for j in range(m + 1)] for i in range(n + 1)]

    for i in range(1, n + 1):
        for j in range(1, m + 1):
            if str1[i - 1] == str2[j - 1]:
                d = 0
            else:
                d = 1

            matrix[i][j] = min(matrix[i - 1][j] + 1, matrix[i][j - 1] + 1, matrix[i - 1][j - 1] + d)

    distance_score = matrix[n][m]

    return distance_score


def url_fix(s: str, encoding: str = "utf-8", remove_fragment: bool = True, remove_query: bool = True) -> str:
    r"""
    Inspired by https://github.com/pallets/werkzeug/blob/main/src/werkzeug/urls.py

    Sometimes you get an URL by a user that just isn't a real URL because
    it contains unsafe characters like ' ' and so on. This function can fix
    some of the problems in a similar way browsers handle data entered by the
    user:
    >>> url_fix('http://de.wikipedia.org/wiki/Elf (Begriffskl\xe4rung)')
    'http://de.wikipedia.org/wiki/Elf%20(Begriffskl%C3%A4rung)'
    :param s: the string with the URL to fix.
    :param encoding: The target encoding for the URL if the url was given
        as a string.
    :param remove_fragment: whether to remove the fragment of the url (the part after '#')
    :return: the fixed urls as a str
    """

    # First step is to switch to text processing and to convert
    # backslashes (which are invalid in URLs anyways) to slashes.  This is
    # consistent with what Chrome does.
    s = str(s).replace("\\", "/")

    url = urlsplit(s)
    path = quote(url.path, safe="/%+$!*'(),", encoding=encoding)
    if remove_fragment:
        fragment = ""
    else:
        fragment = quote_plus(url.fragment, safe=":&%=+$!*'(),", encoding=encoding)

    if remove_query:
        query = ""
    else:
        query = quote_plus(url.query, safe=":&%=+$!*'(),", encoding=encoding)

    return urlunsplit((url.scheme, url.netloc, path, query, fragment))


def replace_quotes(s: str) -> str:
    """replace dificult characters"""
    return s.translate(str.maketrans("‘’“”", "''\"\""))


def replace_accented_letters(s: str) -> str:
    """replace accented letters with their ascii counterpart"""
    return unicodedata.normalize("NFD", s).encode("ascii", "ignore").decode("utf-8")


def init_nlp():
    """Initialize the NLP pipeline"""

    # check whether any language model is installed
    if len(spacy.util.get_installed_models()) == 0:
        log.error("No language model found. Install with ´python -m spacy download es_core_news_sm´")
        raise RuntimeError

    # find closest match to model in cfg
    # via minimum edit distance to installed models
    m_cfg = general_config["lang_model"]

    model_match = sorted([(m, med(m_cfg, m)) for m in spacy.util.get_installed_models()], key=lambda t: t[1])[0][0]

    log.info(f"Chosing {model_match}")
    if model_match[:2] != general_config["lang_model"][:2]:
        log.warning(f"language mismatch: {model_match} != {general_config['lang_model']}1")
    _nlp = spacy.load(model_match)
    language = general_config["lang_model"][:2]
    _nlp.add_pipe("syllables", after="morphologizer", config={"lang": language})

    return _nlp


def init_language_tool(locale):
    """Initializs Languagetool"""
    import language_tool_python

    language_tool = language_tool_python.LanguageTool(locale[:2])
    return language_tool


def get_bag_of_words(doc: spacy.tokens.doc.Doc) -> tuple[set]:
    """generate a bag of normalized words from text"""

    # without_stopwords = {word.text for word in doc if word.is_alpha and not word.is_stop}
    normalized_bow = sorted(
        {word.lemma_.lower() for word in doc if word.is_alpha and not word.is_stop and len(word) > 1}
    )
    places = sorted({word.text.title().strip() for word in doc.ents if word.label_ in ["GPE", "LOC"]})
    people = sorted({word.text.title().strip() for word in doc.ents if word.label_ == "PER"})
    orgs = sorted({word.text.strip() for word in doc.ents if word.label_ == "ORG"})

    # add GPEs from files
    gpes_from_files = []
    for f in glob(os.path.join(general_config["gpe_directory"], "gpe_*.txt")):
        with open(f) as fp:
            gpes_from_files += [line.strip() for line in fp.readlines()]

    gpes_from_files = set(gpes_from_files)

    found_gpes = [word.title() for word in normalized_bow if word.lower() in map(str.lower, gpes_from_files)]
    places = sorted(set(places + found_gpes))

    # fix some frequent errors
    if "Mich." in people:
        people.remove("Mich.")
        places = sorted(set(places + ["Michoacán"]))

    if "MORELIA" in orgs:
        orgs.remove("MORELIA")
        places = sorted(set(places + ["Morelia"]))

    orgs_dict = {
        "Instituto Mexicano del Seguro Social": "IMSS",
        "Secretaría de Seguridad Pública": "SSP",
        "Universidad Michoacana de San Nicolás de Hidalgo": "UMSNH",
    }

    for k, v in orgs_dict.items():
        if k in orgs:
            orgs.remove(k)
            places = sorted(set(places + [v]))

    for p in people.copy():
        if "\n" in p:
            people.remove(p)
            people = sorted(set(people + [p.split("\n")[0]]))

    for p1 in people.copy():
        for p2 in people.copy():
            if p1 != p2 and (p2.startswith(p1) or p2.endswith(p1)) and p1 in people:
                people.remove(p1)

    for p in places.copy():
        if "\n" in p:
            places.remove(p)
            places = sorted(set(places + [p.split("\n")[0]]))

    # print(text)
    # print(without_stopwords)
    # print(normalized_bow)
    # print(places)
    # print(people)
    # for ent in doc.ents:
    #     print(ent.text, ent.label_)
    # sys.exit(999)

    return normalized_bow, places, people, orgs


def process_article_text(
    article_dict: dict,
    config: dict,
    nlp,
    language_tool=None,
    i: int = None,
) -> None:
    """analyse the article text and add the information to the article dict"""

    if len(article_dict["text"]) < 100:
        log.warning(f"{i} found very short text")
        return

    for field in ["text", "subtitle", "title"]:
        if article_dict[field]:
            article_dict[field] = re.sub("\xa0", " ", article_dict[field])
            article_dict[field] = re.sub(r" +", " ", article_dict[field])
            article_dict[field] = re.sub(r"\n\n+", "\n", article_dict[field])
            article_dict[field] = re.sub(r"( *\n *)+", "\n", article_dict[field])

    text_doc = nlp(article_dict["text"])

    # get bag of words
    (
        article_dict["bag_of_words"],
        article_dict["places"],
        article_dict["people"],
        article_dict["orgs"],
    ) = get_bag_of_words(text_doc)

    # get legibility and stats
    article_dict["legibility"] = legibility_es(text_doc)

    article_dict["stats"] = {}
    article_dict["stats"]["n_characters"] = len(article_dict["text"])
    article_dict["stats"]["n_letters"] = count_letters(text_doc)
    article_dict["stats"]["n_syllables"] = count_syllables(text_doc)
    article_dict["stats"]["n_words"] = count_words(text_doc)
    article_dict["stats"]["n_sentences"] = count_sentences(text_doc)
    article_dict["stats"]["n_bag_of_words"] = len(article_dict["bag_of_words"])

    if language_tool is not None:
        matches = language_tool.check(article_dict["text"])
        article_dict["n_errors"] = len(matches)
    else:
        article_dict["n_errors"] = None

    sanitize_article(article_dict)


def sanitize_article(article_dict: dict) -> None:
    """sanitize (i.e., normalize) the keywords, people, categories, etc."""

    # create missing fields
    if "category" not in article_dict:
        article_dict["category"] = []
    if "places" not in article_dict:
        article_dict["places"] = []
    if "authors" not in article_dict:
        article_dict["authors"] = []
    if "keywords" not in article_dict:
        article_dict["keywords"] = []
    if "orgs" not in article_dict:
        article_dict["orgs"] = []
    if "people" not in article_dict:
        article_dict["people"] = []

    # sanitize dates
    # if article_dict["published"].endswith("Z"):
    #     article_dict["published"] = article_dict["published"][:-1]

    tz_correction = timedelta(0)
    if article_dict["outlet"] in ["urbistv", "ultranoticias", "Ecos de La Meseta", "Acueducto Online"]:
        tz_correction = timedelta(hours=-6)

    if article_dict["url"].startswith("https://www.debate.com.mx/policiacas"):
        if not re.search(r"[+-]0\d:00$", article_dict["published"]):
            article_dict["published"] = article_dict["published"] + "+00:00"

    article_dict["published"] = (
        (datetime.fromtimestamp(datetime.fromisoformat(article_dict["published"]).timestamp()) + tz_correction)
        .astimezone(tz=ZoneInfo("America/Mexico_City"))
        .isoformat(timespec="minutes")
    )

    # sanitize authors
    article_dict["authors"] = [key.strip() for key in article_dict["authors"]]
    article_dict["authors"] = [key for key in article_dict["authors"] if len(key) < 100]

    if "Europa Press, Afp" in article_dict["authors"]:
        article_dict["authors"].remove("Europa Press, Afp")
        article_dict["authors"].append("Europa Press")
        article_dict["authors"].append("AFP")

    for key in ["Afp", "AFp", "afp"]:
        if key in article_dict["authors"]:
            article_dict["authors"].remove(key)
            article_dict["authors"] = sorted(set(article_dict["authors"] + ["AFP"]))

    # sanitize categories
    article_dict["category"] = [key.lower() for key in article_dict["category"]]

    # sanitize keywords
    if len(article_dict["keywords"]) == 1 and "," in article_dict["keywords"]:
        article_dict["keywords"] = [key.strip() for key in article_dict["keywords"][0].split(",")]
    article_dict["keywords"] = [key.lower() for key in article_dict["keywords"]]
    for key in ["amlo-presidente", "lopez-obrador", "noticias-amlo"]:
        if key in article_dict["keywords"]:
            article_dict["keywords"].remove(key)
            article_dict["keywords"] = sorted(set(article_dict["keywords"] + ["amlo"]))
    for key in ["corona", "coronavirus", "covid 19"]:
        if key in article_dict["keywords"]:
            article_dict["keywords"].remove(key)
            article_dict["keywords"] = sorted(set(article_dict["keywords"] + ["covid-19"]))
    for key in ["Covid-19", "COVID-19", "COVID-19.", "Covid", "COVID"]:
        if key in article_dict["places"]:
            article_dict["places"].remove(key)
            article_dict["keywords"] = sorted(set(article_dict["keywords"] + ["covid-19"]))
        if key in article_dict["people"]:
            article_dict["people"].remove(key)
            article_dict["keywords"] = sorted(set(article_dict["keywords"] + ["covid-19"]))
        if key in article_dict["orgs"]:
            article_dict["orgs"].remove(key)
            article_dict["keywords"] = sorted(set(article_dict["keywords"] + ["covid-19"]))

    # sanitize people
    for key in article_dict["people"].copy():
        if "   " in key:
            article_dict["people"].remove(key)
            article_dict["people"] = sorted(set(article_dict["people"] + [k.strip() for k in key.split("   ")]))

    article_dict["people"] = [key.title() for key in article_dict["people"]]
    article_dict["people"] = [key.strip(".") for key in article_dict["people"]]
    article_dict["people"] = [key.split(":")[0] for key in article_dict["people"]]
    for key in ["Amlo", "López Obrador"]:
        if key in article_dict["people"]:
            article_dict["people"].remove(key)
            article_dict["people"] = sorted(set(article_dict["people"] + ["Andrés Manuel López Obrador"]))
    for key in ["Silvano Aureoles"]:
        if key in article_dict["people"]:
            article_dict["people"].remove(key)
            article_dict["people"] = sorted(set(article_dict["people"] + ["Silvano Aureoles Conejo"]))
    for key in ["Ramírez Bedolla"]:
        if key in article_dict["people"]:
            article_dict["people"].remove(key)
            article_dict["people"] = sorted(set(article_dict["people"] + ["Alfredo Ramírez Bedolla"]))
    for key in ["Recomendamos El Podcast", "Él", "Instagram"]:
        if key in article_dict["people"]:
            article_dict["people"].remove(key)

    # sanitize places
    article_dict["places"] = [key.strip(".") for key in article_dict["places"]]
    for key in ["Mexico"]:
        if key in article_dict["places"]:
            article_dict["places"].remove(key)
            article_dict["places"] = sorted(set(article_dict["places"] + ["México"]))
    for key in ["En Estados Unidos"]:
        if key in article_dict["places"]:
            article_dict["places"].remove(key)
            article_dict["places"] = sorted(set(article_dict["places"] + ["Estados Unidos"]))
    for key in ["Mich", "Michoacana", "Michoacan"]:
        if key in article_dict["places"]:
            article_dict["places"].remove(key)
            article_dict["places"] = sorted(set(article_dict["places"] + ["Michoacán"]))

    for key in ["Escucha El Podcast ⬇️", "Ver", "Changoonga.Com", "Mascota"]:
        if key in article_dict["places"]:
            article_dict["places"].remove(key)

    # sanitize orgs
    article_dict["orgs"] = [key.strip(".-") for key in article_dict["orgs"]]
    for key in ["Michoacán"]:
        if key in article_dict["orgs"]:
            article_dict["orgs"].remove(key)
    for key in ["Morena"]:
        if key in article_dict["people"]:
            article_dict["people"].remove(key)
            article_dict["orgs"] = sorted(set(article_dict["orgs"] + ["Morena"]))

    for long_name, short_name in (
        ("Secretaría de Salud de Michoacán", "SSM"),
        ("Instituto Nacional Electoral", "INE"),
        ("Organización Mundial de la Salud", "OMS"),
        ("Comisión Federal de Electricidad", "CFE"),
        ("Fiscalía General de la República", "FGR"),
        ("Fiscalía General del Estado", "FGE"),
        ("Partido del Trabajo", "PT"),
    ):
        if long_name in article_dict["orgs"]:
            article_dict["orgs"].remove(long_name)
            article_dict["orgs"] = sorted(set(article_dict["orgs"] + [short_name]))


def sanitize_filename(s: str) -> str:
    """replace characters that should not be in a filename"""

    result = s.translate(str.maketrans("áéíóúñÁÉÍÓÚÑü", "aeiounAEIOUNu", ",\"'“”|.;/‘’?¿!¡&#\t\r\b\n"))

    if len(result) > 100:
        print(result)
        while "___" in result:
            result = result.replace("___", "__")

    return result


def count_syllables(doc: spacy.tokens.doc.Doc) -> int:
    """Count syllables in text"""
    # logging.debug(str([(token.text, token._.syllables_count) for token in doc if not token.is_punct]))
    return sum(token._.syllables_count for token in doc if not token.is_punct and token._.syllables_count is not None)


def count_letters(doc: spacy.tokens.doc.Doc) -> int:
    """Count letters in text"""
    return sum(len(word) for word in doc if not word.is_punct)


def count_words(doc: spacy.tokens.doc.Doc) -> int:
    """Count words in text"""
    return len(list(word for word in doc if not word.is_punct))


def count_sentences(doc: spacy.tokens.doc.Doc) -> int:
    """Count sentences in text"""
    return len(list(doc.sents))


def legibility_es(doc: spacy.tokens.doc.Doc) -> dict:
    """Get legibility measures for Spanish text"""
    result = {}

    n_letters = count_letters(doc)
    n_syllables = count_syllables(doc)
    n_words = count_words(doc)
    n_sentences = count_sentences(doc)
    result["FH"] = result["Fernández Huerta"] = round(
        206.84 - 60 * n_syllables / n_words - 102 * n_sentences / n_words, 1
    )
    result["SP"] = result["Szigriszt-Pazos"] = round(206.835 - 62.3 * n_syllables / n_words - n_words / n_sentences, 1)
    letters_per_word = [len(word) for word in doc if not word.is_punct]
    result["mu"] = round(
        n_words / (n_words - 1) * (n_letters / n_words) / np.var(letters_per_word) * 100,
        1,
    )

    return result


if __name__ == "__main__":
    nlp = spacy.load("es_core_news_lg")
    nlp.add_pipe("syllables", after="morphologizer", config={"lang": "es"})

    assert count_letters(nlp("Esto es un texto.")) == 13
    assert count_syllables(nlp("Esto es un texto.")) == 6
    assert count_words(nlp("Esto es un texto.")) == 4
    assert count_sentences(nlp("Esto es un texto. Y esto es otro.")) == 2

    print(legibility_es(nlp("Esto es un texto muy simple.")))

    long_text = nlp(
        "Tomando en cuenta la complejidad del vocabulario,"
        " esto se podría considerar un texto más complicado y menos legible."
    )

    print(
        count_letters(long_text),
        count_syllables(long_text),
        count_words(long_text),
        count_sentences(long_text),
    )
    print(legibility_es(nlp(long_text)))



# isort: skip_file
from .lib.config import general_config, load_general_config
from .lib.log import console, log, logging_set_up, print_logo, set_log_level, profile, debug, timer_start, timer_stop
from .lib.db import init_mongodb, save_article_to_mongodb, get_articles_cache
from .lib.helpers import memoize
from .lib.text import (
    get_bag_of_words,
    init_language_tool,
    init_nlp,
    preprocess,
    process_article_text,
    sanitize_article,
    sanitize_filename,
    url_fix,
)
from .lib.topics import get_topic_title, hdp_pipeline, init_openai



"""The general configuration"""

import os

import toml

general_config = {
    "_MONITO_DEBUG_MODE": False,
    "_MONITO_QUIET_MODE": False,
}


def load_general_config(args_general_config: str) -> list:
    """load the general configuration"""

    log_msgs = []
    if os.path.isfile(args_general_config):
        log_msgs.append(("DEBUG", f"loading general config file {args_general_config}"))
        # g_config = toml.load(args_general_config)

        try:
            g_config = toml.load(args_general_config)
        except Exception as e:
            log_msgs.append(("ERROR", f"Error loading {args_general_config}:\n{str(e)}"))
            g_config = {
                "log_to_file": "monito_$DATE.log",
                "log_to_file_width": 120,
                "lang_model": "es_core_news_lg",
                "db": "mongodb://localhost:27017",
            }
            general_config.update(g_config)
            return log_msgs
    else:
        g_config = {
            "log_to_file": "monito_$DATE.log",
            "log_to_file_width": 120,
            "lang_model": "es_core_news_lg",
            "db": "mongodb://localhost:27017",
        }
        log_msgs.append(
            (
                "WARNING",
                f"general config file {args_general_config} not found. Falling back to defaults.",
            )
        )

    g_config["hide_progress"] = False

    general_config.update(g_config)

    return log_msgs



"""Find articles similar to a given text file
"""
import argparse
import os
import pickle
import sys
from datetime import datetime
from typing import Dict, List

import matplotlib.pyplot as plt
import numpy as np
import pymongo
from pymongo import MongoClient
from scipy import stats
from sklearn.decomposition import PCA, TruncatedSVD
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity

from monito_tools import (
    console,
    general_config,
    init_mongodb,
    init_nlp,
    load_general_config,
    log,
    logging_set_up,
    preprocess,
    print_logo,
    set_log_level,
)

N_DEBUG_ARTICLES = 300


# our labeling
LABELS: List[str] = [
    "UNLABELED",
    "politics",
    "local",
    "health",
    "news",
    "sports",
    "security",
    "science",
    "technology",
    "environment",
    "entertainment",
    "economics",
    "local",
    "cdmx",
    "euo",
    "ambiente",
    "culture",
]

# spanish categories provided by websites -> our labels
LABEL_LOOKUP: Dict[str, str] = {
    "UNLABELED": "UNLABELED",
    "policiacar": "politics",
    "consejo": "politics",
    "gobierno": "politics",
    "migracion": "politics",
    "justicia": "politics",
    "morelia": "local",
    "deporte": "sports",
    "mazatlan": "local",
    "salud": "health",
    "seguridad": "security",
    "show": "entertainment",
    "entretenimiento": "entertainment",
    "espectaculo": "entertainment",
    "bolsillo": "entertainment",
    "rojo": "entertainment",
    "cultura": "culture",
    "sociedad": "culture",
    "postura": "culture",
    "politico": "politics",
    "noticia": "news",
    "michoacan": "local",
    "destacado": "news",
    "vandalismo": "news",
    "ciencia": "science",
    "viral": "news",
    "astronomia": "science",
    "tecnologia": "technology",
    "principal": "news",
    "ambiente": "environment",
    "hardnews": "news",
    "mexico": "local",
    "patzcuaro": "local",
    "cdmx": "cdmx",
    "euo": "euo",
    "economia": "economics",
    "zitacuarir": "local",
    "palencia": "local",
    "local": "local",
    "estatal": "local",
    "hora": "news",
    "suceso": "news",
    "mundo": "news",
    "guadalajara": "local",
}


def vae_pipeline(X_train: np.ndarray, Y_train: np.ndarray, X_test: np.ndarray, Y_test: np.ndarray, n_dims=3):
    import torch
    from torch import nn
    from torch.optim import Adam
    from torch.utils.data import DataLoader, TensorDataset

    # Define the Variational Autoencoder
    class VAE(nn.Module):
        def __init__(self, input_shape: int, n_labels: int):
            super(VAE, self).__init__()
            self.encoder = nn.Sequential(nn.Linear(input_shape, 128), nn.ReLU(), nn.Linear(128, 64), nn.ReLU())

            self.mu = nn.Linear(64, n_dims)
            self.log_var = nn.Linear(64, n_dims)

            self.decoder = nn.Sequential(
                nn.Linear(n_dims, 32),
                nn.ReLU(),
                nn.Linear(32, 32),
                nn.ReLU(),
                nn.Linear(32, n_labels),
                nn.Softmax(dim=1),
            )

        def encode(self, x):
            h = self.encoder(x)
            mu = self.mu(h)
            log_var = self.log_var(h)
            # reparameterize
            std = torch.exp(log_var / 2)
            eps = torch.randn_like(std)
            z = mu + eps * std
            return z, mu, log_var

        def decode(self, z):
            return self.decoder(z)

        def forward(self, x):
            z, mu, log_var = self.encode(x)
            predictions = self.decode(z)
            return predictions, mu, log_var

    # Convert your data to PyTorch tensors
    X_train_tensor = torch.tensor(X_train, dtype=torch.float32)
    # Y_train_tensor = torch.tensor(Y_train, dtype=torch.float32)
    Y_train_tensor = torch.tensor(np.argmax(Y_train, axis=1), dtype=torch.long)

    # Create a DataLoader
    train_dataset = TensorDataset(X_train_tensor, Y_train_tensor)
    train_dataloader = DataLoader(train_dataset, batch_size=32, shuffle=True)

    # Create the autoencoder, the optimizer and the loss function
    model = VAE(input_shape=X_train_tensor.shape[1], n_labels=len(LABELS))
    optimizer = Adam(model.parameters(), lr=1e-3)

    def loss_fn(predictions, labels, mu, log_var, gamma=1.0):
        CE = nn.functional.cross_entropy(predictions, labels, reduction="sum")
        KLD = -0.5 * torch.sum(1 + log_var - mu.pow(2) - log_var.exp())
        return CE + gamma * KLD

    # Train the autoencoder
    epochs = 20
    print(f"Training an AutoEncoder for {epochs} epochs")

    gamma = 1e-10
    for epoch in range(epochs):
        total_loss = 0
        correct_predictions = 0
        total_predictions = 0
        for batch in train_dataloader:
            x, y = batch
            optimizer.zero_grad()
            y_pred, mu, log_var = model(x)
            loss = loss_fn(y_pred, y, mu, log_var, gamma=gamma)
            loss.backward()
            optimizer.step()

            # Calculate accuracy
            _, predicted = torch.max(y_pred.data, 1)
            total_predictions += y.size(0)
            correct_predictions += (predicted == y).sum().item()

            total_loss += loss.item()

        gamma *= 10.0
        accuracy = correct_predictions / total_predictions * 100
        print(f"Epoch: {epoch}, Loss: {total_loss / total_predictions}, Accuracy: {accuracy}%")

    # to visualize z of train set
    representations_train = model.encode(X_train_tensor)[0].detach().numpy()

    # ------ run inference on test data ------

    X_test_tensor = torch.tensor(X_test, dtype=torch.float32)
    Y_test_tensor = torch.tensor(np.argmax(Y_test, axis=1), dtype=torch.long)

    representations_test = model.encode(X_test_tensor)[0].detach().numpy()
    predictions_test = model(X_test_tensor)[0].detach().numpy()

    test_dataset = TensorDataset(X_test_tensor, Y_test_tensor)
    test_dataloader = DataLoader(test_dataset, batch_size=32, shuffle=True)

    # predict on test data
    test_correct_predictions = 0
    test_total_predictions = 0
    for batch in test_dataloader:
        x, y = batch
        y_pred, mu, log_var = model(x)

        _, predicted = torch.max(y_pred.data, 1)
        test_correct_predictions += (predicted == y).sum().item()
        test_total_predictions += y.size(0)

    test_accuracy = test_correct_predictions / test_total_predictions * 100
    print(f"Test accuracy: {test_accuracy:.3f}%")

    return representations_train, representations_test, predictions_test


def reduce_dimensionality(
    X,  # array of shape N x D
    method="svd",
    n_dims=3,
):
    assert method in ["svd", "pca", "ae", "vae"]
    if method == "svd":
        svd = TruncatedSVD(n_components=n_dims)
        X_reduced = svd.fit_transform(X)
    elif method == "pca":
        pca = PCA(n_components=n_dims)
        X_reduced = pca.fit_transform(X)
    elif method == "ae":
        import torch
        from torch import nn
        from torch.optim import Adam
        from torch.utils.data import DataLoader, TensorDataset

        # Define the Autoencoder
        class Autoencoder(nn.Module):
            def __init__(self):
                super(Autoencoder, self).__init__()
                D = X.shape[1]
                self.encoder = nn.Sequential(
                    nn.Linear(D, 128), nn.ReLU(), nn.Linear(128, 64), nn.ReLU(), nn.Linear(64, n_dims)
                )
                self.decoder = nn.Sequential(
                    nn.Linear(n_dims, 64), nn.ReLU(), nn.Linear(64, 128), nn.ReLU(), nn.Linear(128, D), nn.Sigmoid()
                )  # use a sigmoid for binary features, remove for continuous features

            def forward(self, x):
                x = self.encoder(x)
                x = self.decoder(x)
                return x

        # Convert your data to PyTorch tensors
        X_train_tensor = torch.tensor(X.toarray(), dtype=torch.float32)

        # Create a DataLoader
        dataset = TensorDataset(X_train_tensor)
        dataloader = DataLoader(dataset, batch_size=32, shuffle=True)

        # Create the autoencoder, the optimizer and the loss function
        model = Autoencoder()
        optimizer = Adam(model.parameters(), lr=1e-3)
        loss_fn = nn.MSELoss()

        # Train the autoencoder
        epochs = 100
        print(f"Training an AutoEncoder for {epochs} epochs")

        for epoch in range(epochs):
            for batch in dataloader:
                x = batch[0]
                optimizer.zero_grad()
                output = model(x)
                loss = loss_fn(output, x)
                loss.backward()
                optimizer.step()
            if epoch % 10 == 0:  # print loss every 10 epochs
                print(f"Epoch: {epoch}, Loss: {loss.item()}")

        # Use the encoder to reduce dimensionality
        X_reduced = model.encoder(X_train_tensor).detach().numpy()

    elif method == "vae":
        raise NotImplemented
    else:
        raise NotImplemented
    return X_reduced


def prepare_corpus(
    cursor: pymongo.cursor.Cursor,
    nlp_model,
    debug=False,
):
    unique_labels: List[str] = []
    labels_np: List[np.ndarray] = []
    labels_int: List[int] = []
    corpus: List[str] = []

    n_labels = len(LABELS)

    for j, c in enumerate(cursor):
        # print(f"Processing train doc",j)
        if j > N_DEBUG_ARTICLES and debug:
            break
        string_label = "UNLABELED"
        categories = c.get("category", [])

        categories = [preprocess(nlp_model, cat) for cat in categories]
        print(categories)
        for cat in categories:
            if not type(cat) == str:
                nested_list = False
                try:
                    if type(cat[0]) == str:
                        nested_list = True
                        for c_ in cat:
                            string_label = c_
                            unique_labels.append(c_)
                except IndexError:
                    continue
            else:
                string_label = cat
            if cat not in unique_labels and not nested_list:
                unique_labels.append(cat)
            nested_list = False

        # get the train label for this article (an index of LABELS)
        string_lbl = string_label.strip()
        lbl = LABEL_LOOKUP.get(string_lbl, "UNLABELED")
        # print(f"During labeling of train data, found a new category '{string_lbl}' that must be manually added to monito.similarity.monito_similarity.LABEL_LOOKUP!")
        lbl = LABELS.index(lbl)
        labels_int.append(lbl)
        label_numpy = np.zeros(n_labels)
        # NOTE: only consider one label per datapoint for now
        label_numpy[lbl] = 1  # TODO add multilabels!!! FIXME

        labels_np.append(label_numpy)

        corpus.append(" ".join(preprocess(nlp_model, c["title"] + "\n" + c["text"])))
    print(f"Unique labels: {unique_labels}")
    return corpus, labels_np, labels_int


def classify(
    cursor_train: pymongo.cursor.Cursor,
    cursor_test: pymongo.cursor.Cursor,
    nlp_model,  # spacy.lang.es.Spanish
    vectorizer: TfidfVectorizer,
    debug: bool = False,
    dimensionality_reduction: str = "svd",
    n_dims: int = 3,
    outlier_threshold: float = 2.0,
    visualize: bool = False,
):
    """
    Classify unseen data using clustering and label propagation.

    :param cursor_train: mongodb cursor containing the documents the vectorizer was already fitted to.
    :param cursor_test: mongodb cursor containing the documents that are to be labeled
    :param nlp_model: The SpaCy language model to use for preprocessing.
    :param vectorizer: The TfidfVectorizer
    :param debug: If True, only do a short run with N_DEBUG_ARTICLES documents for each of train/test
    :param dimensionality_reduction: One method out of ("svd", "pca", "ae", "vae")
    :param n_dims: number of dimensions to reduce to
    :param outlier_threshold: the number of standard deviations away from the mean from whereon outliers are ignored (in the reduced space)
    """

    train_corpus, train_labels_np, train_labels_int = prepare_corpus(cursor_train, nlp_model, debug=debug)
    train_labels_np = np.stack(train_labels_np)
    test_corpus, test_labels_np, test_labels_int = prepare_corpus(cursor_test, nlp_model, debug=debug)
    test_labels_np = np.stack(test_labels_np)

    X_train = vectorizer.transform(train_corpus).toarray()
    X_test = vectorizer.transform(test_corpus).toarray()

    repr_train, repr_test, predictions = vae_pipeline(X_train, train_labels_np, X_test, test_labels_np, n_dims=n_dims)
    # # Reduce dimensionality
    # X_train_reduced = reduce_dimensionality(
    #     X_train, method=dimensionality_reduction
    # )
    # X_test_reduced = reduce_dimensionality(
    #     X_test, method=dimensionality_reduction
    # )

    # Outlier removal
    # calculate z-scores
    # z_scores = np.abs(stats.zscore(X_train_reduced.numpy()))

    # # remove outliers
    # X_train_reduced = X_train_reduced[
    #     (z_scores < outlier_threshold).all(axis=1)
    # ]

    if visualize:
        fig = plt.figure(figsize=(16, 12))
        ax_train = fig.add_subplot(121, projection="3d")
        ax_test = fig.add_subplot(122, projection="3d")

        scatter_train = ax_train.scatter(repr_train[:, 0], repr_train[:, 1], repr_train[:, 2], c=train_labels_int)
        scatter_test = ax_test.scatter(repr_test[:, 0], repr_test[:, 1], repr_test[:, 2], c=test_labels_int)
        plt.title(f"3D {dimensionality_reduction.upper()} of train, test")
        plt.show()

    # sim_matrix = cosine_similarity(matrix_old, matrix_new)

    # print(sim_matrix)

    # count = len(texts)
    # count_groups = 0
    # already_processed = set()
    # for i in range(count):
    #     printed_first = False
    #     if i in already_processed:
    #         continue
    #     for j in range(count):
    #         if i == j:
    #             continue
    #         if j in already_processed:
    #             continue
    #         if sim_matrix[i][j] > args.sensitivity:
    #             if not printed_first:
    #                 print(f"\n{i}\t{outlets[i]:30} {titles[i]} {urls[i]}")
    #                 count_groups += 1
    #                 printed_first = True
    #             # print(round(sim_matrix[i][j], 3), i, j)
    #             print(f"{j}\t{outlets[j]:30} {titles[j]} {urls[j]}")
    #             already_processed.add(j)
    #     already_processed.add(i)

    # print("\n", count_groups)


def _parse_arguments(argv=None):
    # parse the command line options
    argparser = argparse.ArgumentParser(
        description="monito.similarity: a tool for monitoring and filtering news feeds. "
        "Please have a look at the documentation (https://monito.readthedocs.io/en/latest/) "
        "for further information on how tho use this software.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argparser.add_argument(
        "-g",
        "--general-config",
        type=str,
        default="monito_config.toml",
        help="The general configuration file.",
    )
    argparser.add_argument(
        "-l",
        "--log-to-stdout",
        action="store_true",
        default=False,
        help="Write to stdout instead to the logfile set in the general config file.",
    )
    argparser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Switch off text output except for error messages. This will overwrite -v.",
    )
    argparser.add_argument("--verbose", action="store_true", help="more verbose text output")
    argparser.add_argument(
        "--reduction",
        type=str,
        help="Dimensionality Reduction Technique. One of 'svd', 'pca', 'ae', 'vae'",
        choices=["svd", "pca", "ae", "vae"],
    )
    argparser.add_argument("--n-dims", type=int, help="Dimensionality to code for.", default=3)

    argparser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="switch on debug mode.",
    )
    argparser.add_argument(
        "--since",
        type=str,
        default="2023-04-01",
        help="The start date for vectorizer creation in ISO format",
    )
    argparser.add_argument(
        "--till",
        type=str,
        default="2023-04-06",
        help="The end date for vectorizer creation in ISO format",
    )

    argparser.add_argument(
        "-s",
        "--start",
        type=str,
        default="2023-04-07",
        help="The start date for classification in ISO format",
    )
    argparser.add_argument(
        "-e",
        "--end",
        type=str,
        default="2023-04-08",
        help="The end date for classification in ISO format",
    )
    argparser.add_argument(
        "-f",
        "--file",
        type=str,
        default="data/vectorizer_2023-01-06_35087.pk",
        help="The vectorizer file to use for classification.",
    )
    argparser.add_argument("-y", "--sensitivity", type=float, default=0.99, help="The sensitivity TODO")
    argparser.add_argument(
        "-c", "--classify", action="store_true", default=False, help="Whether to classify new articles"
    )
    argparser.add_argument(
        "-v", "--create-vectorizer", action="store_true", default=False, help="Whether to create a new vectorizer"
    )

    return argparser.parse_args(argv)


def run(argv: list = None):
    """the command line tool. Please use the ``--help`` option to get help."""

    # parse the command line options
    args = _parse_arguments(argv)

    # load the general configuration
    log_msgs = load_general_config(args.general_config)

    # set up the output channel
    logging_set_up(args.log_to_stdout)

    # print log messages from loading the general config file
    # for level, msg in log_msgs:
    #     print(level, msg)
    #     log.log(getattr(logging, level), msg)

    # print the logo and version
    if not args.quiet:  # pragma: no cover
        print_logo()

    # set quiet mode
    general_config["_MONITO_QUIET_MODE"] = args.quiet

    # set debug mode
    general_config["_MONITO_DEBUG_MODE"] = args.debug

    # set verbosity level
    set_log_level(args.quiet, args.verbose, args.debug)

    # init data base
    articles_db = init_mongodb()

    if not (args.create_vectorizer or args.classify):
        print(f"Nothing to do (either --create-vectorizer or --classify must be passed.).")
        exit()

    nlp = init_nlp()

    cursor_old = articles_db.find({"published": {"$gte": args.since, "$lte": args.till}})

    # Create a vectorizer or load it from file
    if args.create_vectorizer:
        print(f"Creating Vectorizer with articles from {args.since} to {args.till}")
        corpus_to_compare_to = [preprocess(nlp, c["title"] + "\n" + c["text"]) for c in cursor_old]

        vectorizer = TfidfVectorizer()
        vectorizer.fit(corpus_to_compare_to)

        filename = f"vectorizer_{datetime.today().strftime('%Y-%m-%d')}_{len(corpus_to_compare_to)}.pk"
        with open(filename, "wb") as f:
            pickle.dump(vectorizer, f)

            print(f"{datetime.today().strftime('[%Y-%m-%d %H:%M:%S]')} saved the vectorizer as {filename}")
            vectorizer_file = filename
        print(f"Saved vectorizer as {filename}.")
    else:
        with open(args.file, "rb") as f:
            print(f"Attempting to load vectorizer from {args.file}")
            vectorizer = pickle.load(f)
            print(f"Successfully loaded vectorizer from {args.file}")

    if args.classify:
        print(f"Running Classification for new articles from {args.start} to {args.end}")
        print(f"by comparing to old articles from {args.since} to {args.till}")
        print(f"using the previously created vectorizer at '{args.file}'")

        cursor_new = articles_db.find({"published": {"$gte": args.start, "$lte": args.end}})

        classify(
            cursor_old,
            cursor_new,
            nlp,
            vectorizer,
            debug=args.debug,
            dimensionality_reduction=args.reduction,
            n_dims=args.n_dims,
        )


if __name__ == "__main__":
    import sys

    try:
        run(sys.argv[1:])
    except KeyboardInterrupt:
        print("Interrupted")
        try:
            sys.exit(2)
        except SystemExit:
            os._exit(2)
    finally:
        if "log_filename" in general_config:
            console.save_html(f"{general_config['log_filename']}.html", clear=False)



import json
import os
import pickle
import sys
import warnings
from datetime import datetime
from typing import Dict, List, Optional, Set, Tuple

from tqdm import tqdm

warnings.filterwarnings("ignore")

import gensim
import nltk
import openai
import pymongo
import spacy
from gensim.corpora import Dictionary
from gensim.models import HdpModel
from pymongo import MongoClient

from monito_tools import general_config, init_mongodb, init_nlp, load_general_config, log, preprocess

MODEL = None
INSTRUCTIONS = None
DEBUG = True


def init_openai():
    global MODEL, INSTRUCTIONS

    openai_api_key: Optional[str] = os.getenv("OPENAI_API_KEY")

    if openai_api_key is None:
        """
        Try reading from file
        """
        openai_filename = ".openai_api_key"
        log.info(f"Attempting to read OpenAI Key from {openai_filename} ...")
        try:
            with open(openai_filename, "r") as api_file:
                openai_api_key = api_file.read()[:-1]
        except FileNotFoundError as FNFE:
            log.warning(f"OpenAI Key must be provided at {openai_filename} if you want to use the topics feature.")
            # raise FNFE
        else:
            log.info(f"Successfully read OpenAI Key from {openai_filename}.")

    openai.api_key = openai_api_key

    MODEL = "gpt-3.5-turbo"
    # MODEL = "gpt-4"

    INSTRUCTIONS = [
        {
            "role": "system",
            "content": "Usted es TopicGPT, una IA que genera un título corto para un tema compuesto por varias palabras.",
        },
        {
            "role": "system",
            "content": "Las palabras son devueltas por un modelo de Proceso Dirichlet Jerárquico.",
        },
        {
            "role": "system",
            "content": "Los títulos deben ser de una sola palabra o sustantivo compuesto. Sea lo más breve posible.",
        },
        {
            "role": "system",
            "content": "Encuentre la mejor palabra que se relacione con todas las palabras clave dadas.",
        },
        {
            "role": "system",
            "content": "Nunca debe emitir más de una palabra o sustantivo compuesto.",
        },
    ]


def get_topic_title(ps_and_ws: List[Tuple[float, str]]) -> str:
    """
    Use GPT to get topic title
    :param ps_and_ws: The words and associated probabilities of one topic returned by hdp_pipeline
    """

    init_openai()

    words = " ".join([w for (p, w) in ps_and_ws])

    input_ = {"role": "user", "content": words}

    try:
        response = openai.ChatCompletion.create(
            model=MODEL,
            messages=INSTRUCTIONS + [input_],
        )

        response = response["choices"][0]["message"]["content"]

        response = response.replace(".", "")
        response = response.replace("'", "")
        response = response.replace('"', "")

        # get last word of response
        response = response.split(" ")[-1]
    except openai.error.RateLimitError:
        response = ""

    return response


def remove_common_words(corpus: List[List[str]], remove_fraction=0.1):
    """
    Remove most frequent words (across documents)
    Does not double count words in the same doc
    :param corpus: List of List of tokens; output of preprocess
    """

    dictionary: Dict[str, int] = {}  # holds wordcounts
    unique_tokens = []

    for doc in corpus:
        seen_in_this_doc: Set[str] = set()
        for token in doc:
            if token not in seen_in_this_doc:
                seen_in_this_doc.add(token)
                if token not in dictionary:
                    dictionary[token] = 1
                    unique_tokens.append(token)
                else:
                    dictionary[token] += 1

    sorted_tokens = sorted(unique_tokens, key=lambda t: dictionary[t], reverse=True)
    remove_n = int(remove_fraction * len(dictionary))
    keep_tokens = set(sorted_tokens[remove_n:])

    cleaned_corpus = []
    for doc in corpus:
        cleaned_corpus.append([])
        for token in doc:
            if token in keep_tokens:
                cleaned_corpus[:-1].append(token)

    return cleaned_corpus


def do_hdp(texts: List[List[str]], **kwargs) -> gensim.models.hdpmodel.HdpModel:
    """
    Build a HDP Model using gensim.
    :param texts: Corpus, i.e. list of list of strings.
    :return hdp_model: The HDP Model
    """
    dictionary = Dictionary(texts)
    corpus = [dictionary.doc2bow(doc) for doc in texts]

    hdp_model = HdpModel(corpus, dictionary, **kwargs)
    return hdp_model


def hdp_pipeline(
    cursor: pymongo.cursor.Cursor,
    num_topics: int = 10,
    num_words: int = 10,
    output_file: Optional[str] = None,
    remove_fraction: float = 0.1,
    **kwargs,
) -> List[Tuple[int, List[Tuple[float, str]]]]:
    """
    Build Topic Model (using gensim's Hierarchical Dirichlet Process)
    from a mongo cursor of a monito article collection
    Example usage:
    >>> from monito import general_config, init_mongodb, load_general_config
    >>> from monito import hdp_pipeline
    >>> load_general_config("monito_config.toml")
    >>> db = init_mongodb()
    >>> cursor = db.find() # add criteria here
    >>> topics = hdp_pipeline(cursor)
    :param cursor: the excerpt of the monito database to analyze
    :param num_topics: the number of most important topics to output
    :param num_words: the number of most important words per topic to output
    :param output_file: if given, path where to save the hdp model
    :param kwargs: keyword arguments of gensim.models.HdpModel
    :return topics: List of topics, numbered and for each a list of words with probability
    """

    nlp = init_nlp()

    # lang_model = general_config["lang_model"]
    # nlp = spacy.load(lang_model)

    corpus = []
    for i, c in tqdm(enumerate(cursor), desc="Preprocessing articles"):
        if DEBUG and i > 100:
            break
        preprocessed = preprocess(nlp, c["title"] + "\n" + c["subtitle"] + "\n" + c["text"])
        corpus.append(preprocessed)

    cleaned_corpus = remove_common_words(corpus, remove_fraction=remove_fraction)

    hdp_model = do_hdp(corpus, **kwargs)

    # print(f"HDP Model methods:")
    # print(dir(hdp_model))

    numbered_topics = hdp_model.print_topics(num_topics, num_words)

    topics = get_topics(numbered_topics)

    if output_file is not None:
        hdp_model.save(output_file)

    return topics


def get_topics(numbered_topics):
    """
    Post process the output of gensim.models.HdpModel.print_topics
    """
    topics = []
    for i, topical in numbered_topics:
        probs_and_words = topical.split(" + ")
        ps_and_ws = []
        for p_w in probs_and_words:
            star = p_w.find("*")
            prob = float(p_w[:star])
            word = p_w[star + 1 :]
            ps_and_ws.append((prob, word))

        topics.append((i, ps_and_ws))
    return topics


def run():
    log_msgs = load_general_config("monito_config.toml")

    # init data base
    articles_db = init_mongodb()

    start = "2023-04-03"
    end = "2023-04-04"

    num_topics = 20
    num_words = 10

    # make corpus
    output_file = "data/topic_model.gensim"

    cursor = articles_db.find({"published": {"$gte": start, "$lte": end}})

    if not os.path.isfile(output_file):
        topics = hdp_pipeline(
            cursor=cursor,
            num_topics=num_topics,
            num_words=num_words,
            output_file=output_file,
        )
    else:
        hdp_model = HdpModel.load(output_file)
        topics = get_topics(hdp_model.print_topics(num_topics, num_words))

    for topic in topics:
        print()
        print(topic)

    # top_p = max([max([p for (p, w) in topic]) for i, topic in topics])


if __name__ == "__main__":
    run()



"""Monito tool for printing general stats"""

from datetime import datetime, timedelta

from rich import print
from rich.panel import Panel
from rich.pretty import Pretty
from rich.table import Table
from rich.text import Text

from monito_tools import init_mongodb, logging_set_up, timer_start, timer_stop


def get_pretty_article(article: dict):
    output_str = Text()
    output_str.append(f"{article['title']:<120}", style="bold")
    output_str.append(f"   {article['outlet']}")
    output_str.append("\n")
    output_str.append(
        f"{article['published']} (saved: {article['saved']} | {article['download_hostname']})\n", style="dim"
    )

    return output_str


class PrettyArticleShort(Pretty):
    def __init__(self):
        super().__init__(self)


if __name__ == "__main__":
    db_articles = init_mongodb()
    logging_set_up()

    timer_start("stats")
    proj = {"title": True, "url": True, "published": True, "saved": True, "outlet": True, "download_hostname": True}

    now = datetime.now()
    now_24h = now - timedelta(seconds=86400)
    now_1w = now - timedelta(days=7)

    cursor = db_articles.find({}, proj, sort=[("published", -1)]).limit(1)
    last_published = list(cursor)[0]

    cursor = db_articles.find({}, proj, sort=[("saved_iso", -1)]).limit(1)
    last_saved = list(cursor)[0]

    general_stats = Table("key", "value", title="General Statistics", width=50)

    general_stats.add_row("total articles", str(db_articles.estimated_document_count()))
    general_stats.add_row("last 24h", str(db_articles.count_documents({"published": {"$gte": now_24h.isoformat()}})))
    general_stats.add_row("last week", str(db_articles.count_documents({"published": {"$gte": now_1w.isoformat()}})))
    print(general_stats)

    print(Panel(get_pretty_article(last_published), title="last published", title_align="left"))
    print(Panel(get_pretty_article(last_saved), title="last saved", title_align="left"))
    timer_stop("stats")



import functools


def memoize(fn=None, **kwargs):
    """
    Caching wrapper that makes a function store
    previous arguments-return pairs in a dict
    and returns the appropriate value if the combination of
    arguments has been seen previously
    """
    if fn is None:
        return lambda _fn: memoize(_fn, **kwargs)
    return functools.lru_cache(**kwargs)(fn)



"""Analysis tools for monito data files"""

__version__ = "0.1.0"

import argparse
import json
import logging
from collections import Counter
from datetime import date, datetime, timedelta
from glob import glob

from fs import open_fs
from fs.path import basename
from rich import box, print

# from rich.color import Color
from rich.columns import Columns

# from rich.console import Console
from rich.logging import RichHandler
from rich.panel import Panel
from rich.progress import track
from rich.table import Table
from rich.text import Text

from monito_tools import console, init_mongodb, load_general_config, log, sanitize_article, set_log_level

# from uniplot import plot

logging.basicConfig(
    level="INFO",
    format="%(message)s",
    datefmt="[%F %T]",
    # handlers=[RichHandler(console=Console(highlighter=IsoformatDateHighlighter(),
    #                                       theme=Theme({"monito.isodate": "bold magenta"})),
    #                       rich_tracebacks=True)],
    handlers=[RichHandler(rich_tracebacks=True, console=console, show_path=False, show_time=False)],
)


colors = [
    "yellow",
    "cyan",
    "magenta",
    "green",
    "red",
    "blue",
]

_MONITO_DEBUG_MODE = False
_MONITO_QUIET_MODE = False
_MONITO_VERBOSE_MODE = False


def _parse_arguments(argv=None):
    # parse the command line options
    argparser = argparse.ArgumentParser(
        description="monito_analysis: a tool for analysing news feeds downloaded by monito. "
        "Please have a look at the documentation (https://monito.readthedocs.io/en/latest/) "
        "for further information on how tho use this software.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    argparser.add_argument(
        "-i",
        "--input-directory",
        type=str,
        nargs="+",
        help="The input directory or directories where the articles are read from. "
        "If absent, the program will connect to mongodb.",
    )
    argparser.add_argument(
        "-o",
        "--output-file",
        type=str,
        default="out.txt",
        help="The output file for results.",
    )
    argparser.add_argument(
        "--outlet",
        type=str,
        nargs="+",
        help="Filter articles by outlets.",
    )
    argparser.add_argument(
        "-s",
        "--save",
        type=str,
        nargs="+",
        help="Which parts of the data to write to the output file.",
    )
    argparser.add_argument(
        "--category",
        type=str,
        nargs="+",
        help="Filter articles by categories.",
    )
    argparser.add_argument(
        "--keyword",
        type=str,
        nargs="+",
        help="Filter articles by keywords.",
    )
    argparser.add_argument(
        "--author",
        type=str,
        nargs="+",
        help="Filter articles by authors.",
    )
    argparser.add_argument(
        "--place",
        type=str,
        nargs="+",
        help="Filter articles by places.",
    )
    argparser.add_argument(
        "--people",
        type=str,
        nargs="+",
        help="Filter articles by people.",
    )
    argparser.add_argument(
        "--org",
        type=str,
        nargs="+",
        help="Filter articles by orgs.",
    )
    argparser.add_argument(
        "--full-text",
        type=str,
        default=None,
        help="Filter articles by full text search.",
    )
    argparser.add_argument(
        "-l",
        "--last",
        type=str,
        metavar="TIME",
        help="Filter articles published since TIME. Specify the TIME in the form n[n...]{h,d,w}, e.g., "
        "'24h', or '1w'",
    )
    argparser.add_argument(
        "-p",
        "--print",
        type=str,
        nargs="+",
        metavar="TOPIC",
        default=["stats"],
        help="Print individual articles. Can be 'stats, 'title', 'text' or 'fulltext'. "
        "'text' and 'fulltext' include 'title'.",
    )
    argparser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Switch off text output except for error messages. This will overwrite -v.",
    )
    argparser.add_argument("-v", "--verbose", action="store_true", help="More verbose text output")
    argparser.add_argument(
        "-d",
        "--debug",
        action="store_true",
        help="Switch on debug mode. This will show intermediate results and plots, as well as "
        "log a lot of debugging information.",
    )
    argparser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
        help="show the version of this software",
    )
    return argparser.parse_args(argv)


def bar(fraction: float, total: int) -> str:
    """return a string containing a bar of the fraction ([0.0, 1.0])
    of the total width using unicode block characters"""

    blocks = (" ", "▏", "▎", "▍", "▌", "▋", "▊", "▉")
    blocks_to_fill = total * fraction
    residual = (blocks_to_fill - int(blocks_to_fill)) * 8
    result = "█" * int(blocks_to_fill) + blocks[int(residual)]
    return result.strip()


def filter_by_outlet(article_dict: dict, args_outlet: list) -> bool:
    if args_outlet:
        for outlet in args_outlet:
            if outlet.lower() in article_dict["outlet"].lower():
                break
        else:
            return True
    return False


def prefilter_by_date(article_filename: str, args_last: str) -> bool:
    """Filter by the date in the filename, rounded up one day"""

    if args_last:
        last_val = int(args_last[:-1])
        last_unit = args_last[-1]
        if last_unit not in ["h", "d", "w"]:
            log.error("valid units for time differences are 'h', 'd', and 'w'")
        article_date = datetime.fromisoformat(article_filename[:10])

        if last_unit == "h":
            td = timedelta(hours=last_val, days=1)
        elif last_unit == "d":
            td = timedelta(days=last_val + 1)
        elif last_unit == "w":
            td = timedelta(weeks=last_val, days=1)

        if article_date < datetime.now() - td:
            return True

    return False


def get_timedelta_from_last_string(last_str: str) -> timedelta:
    last_val = int(last_str[:-1])
    last_unit = last_str[-1]
    if last_unit not in ["h", "d", "w"]:
        log.error("valid units for time differences are 'h', 'd', and 'w'")
        return

    if last_unit == "h":
        td = timedelta(hours=last_val)
    elif last_unit == "d":
        td = timedelta(days=last_val)
    elif last_unit == "w":
        td = timedelta(weeks=last_val)
    return td


def filter_by_date(article_dict: dict, args_last: str) -> bool:
    if len(article_dict["published"]) == 0:
        log.warning(f"empty field 'published' \n {article_dict}")
        return True

    if args_last:
        article_date = datetime.fromisoformat(article_dict["published"][:16])
        td = get_timedelta_from_last_string(args_last)

        if article_date < datetime.now() - td:
            return True

    return False


def filter_by_category(article_dict: dict, args_category: list):
    if args_category:
        for category in args_category:
            if any(category.lower() in cat.lower() for cat in article_dict["category"]):
                break
        else:
            return True
    return False


def filter_by_keyword(article_dict: dict, args_keyword: list):
    if args_keyword:
        for keyword in args_keyword:
            if any(keyword.lower() in art.lower() for art in article_dict["keywords"]):
                break
        else:
            return True
    return False


def filter_by_author(article_dict: dict, args_author: list):
    if args_author:
        for author in args_author:
            if any(author.lower() in aut.lower() for aut in article_dict["authors"]):
                break
        else:
            return True
    return False


def filter_by_place(article_dict: dict, args_place: list):
    if args_place:
        for place in args_place:
            if any(place.lower() in pl.lower() for pl in article_dict["places"]):
                break
        else:
            return True
    return False


def filter_by_people(article_dict: dict, args_people: list):
    if args_people:
        for person in args_people:
            if any(person.lower() in per.lower() for per in article_dict["people"]):
                break
        else:
            return True
    return False


def filter_by_org(article_dict: dict, args_org: list):
    if args_org:
        for org in args_org:
            if any(org.lower() in o.lower() for o in article_dict["orgs"]):
                break
        else:
            return True
    return False


def filter_by_full_text_search(article_dict: dict, args_full_text: list):
    if args_full_text:
        for text in args_full_text:
            if text.lower() in article_dict["text"].lower():
                break
        else:
            return True
    return False


def display_size(size: int) -> str:
    dir_size_KB = round(size / 1024, 1)
    dir_size_MB = round(dir_size_KB / 1024, 1)
    if dir_size_MB > 1:
        display_size = f"{dir_size_MB} MB"
    else:
        display_size = f"{dir_size_KB} KB"

    return display_size


def run(argv: list = None):
    """the command line tool. Please use the ``--help`` option to get help."""

    global _MONITO_QUIET_MODE
    global _MONITO_DEBUG_MODE
    global _MONITO_VERBOSE_MODE

    # parse the command line options
    args = _parse_arguments(argv)

    # print the logo and version
    if args.quiet is False:  # pragma: no cover
        console.print(
            Panel(
                Text(f"monito_analysis {__version__}", justify="center", style="bold red"),
                box=box.DOUBLE_EDGE,
                highlight=True,
            )
        )

    # set quiet mode
    _MONITO_QUIET_MODE = args.quiet

    # set debug mode
    _MONITO_DEBUG_MODE = args.debug

    # set verbose mode
    _MONITO_VERBOSE_MODE = args.verbose

    # set verbosity level
    set_log_level(args.quiet, args.verbose, args.debug)

    if _MONITO_VERBOSE_MODE or _MONITO_DEBUG_MODE:
        log.info(args)

    files = []
    total_size = 0
    total_n = 0

    if args.input_directory:
        for directory in args.input_directory:
            if _MONITO_VERBOSE_MODE:
                log.info(f"processing directory {directory}")
            files_dir = glob(f"{directory}/*.json")
            if _MONITO_VERBOSE_MODE:
                log.info(f"found {len(files_dir)} files.")

            article_fs = open_fs(directory)
            dir_size = sum(info.size for info in list(article_fs.scandir("/", namespaces=["details"])))
            if _MONITO_VERBOSE_MODE:
                log.info(f"size of files in directory: {display_size(dir_size)}")

            total_size += dir_size
            total_n += len(files_dir)
            files.extend(files_dir)

        log.info(f"total number of files {total_n}")

        log.info(f"size of all files: {display_size(total_size)}")

        log.info("processing articles")

        found_articles = []

        for f in track(files, description="reading article files…", transient=True):
            if prefilter_by_date(basename(f), args.last):
                continue

            try:
                article_dict = json.load(open(f))
            except Exception as e:
                logging.error(f"failed to decode file {f}")
                logging.error(e)
                continue

            # filters
            if filter_by_outlet(article_dict, args.outlet):
                continue
            if filter_by_date(article_dict, args.last):
                continue
            if filter_by_category(article_dict, args.category):
                continue
            if filter_by_keyword(article_dict, args.keyword):
                continue
            if filter_by_author(article_dict, args.author):
                continue
            if filter_by_place(article_dict, args.place):
                continue
            if filter_by_people(article_dict, args.people):
                continue
            if filter_by_org(article_dict, args.org):
                continue
            if filter_by_full_text_search(article_dict, args.full_text):
                continue

            found_articles.append(article_dict)

    else:
        log.info("Connecting to MongoDB.")
        # init data base
        load_general_config("monito_config.toml")
        articles_db = init_mongodb()
        and_arr = []
        if args.outlet:
            and_arr.append({"outlet": {"$in": args.outlet}})
        if args.last:
            td = get_timedelta_from_last_string(args.last)
            start_date = (date.today() - td).isoformat()
            and_arr.append({"published": {"$gte": start_date}})
        if args.full_text:
            and_arr.append({"$text": {"$search": args.full_text}})

        query = {"$and": and_arr}
        found_articles = list(articles_db.find(query))

    # checks and extract data
    data = {
        "outlet": [],
        "authors": [],
        "keywords": [],
        "places": [],
        "people": [],
        "orgs": [],
        "category": [],
        "published": [],
    }

    article_count = 0
    for article_dict in found_articles:
        # sanitize
        sanitize_article(article_dict)

        # checks
        if "saved" in article_dict and article_dict["published"] > article_dict["saved"]:
            log.debug(f"'published' is in the future \n {article_dict}")

        # extract data
        for k in data:
            if isinstance(article_dict[k], list):
                data[k].extend(article_dict[k])
            else:
                data[k].append(article_dict[k])

        article_count += 1

    log.info(f"number of articles: {article_count}")

    if article_count == 0:
        return

    data["weekday"] = []
    data["month"] = []
    data["hour"] = []
    data["monthday"] = []
    # data["week"] = []
    data["date"] = []
    data["datemonth"] = []

    weekdays = [
        "Monday",
        "Tuesday",
        "Wednesday",
        "Thursday",
        "Friday",
        "Saturday",
        "Sunday",
    ]

    earliest_pub = "9"
    latest_pub = "0"
    for pub in data["published"]:
        if pub < earliest_pub:
            earliest_pub = pub
        if pub > latest_pub:
            latest_pub = pub
        pub_dt = datetime.fromisoformat(pub)
        data["weekday"].append(weekdays[pub_dt.weekday()])
        data["month"].append(pub_dt.month)
        data["hour"].append(pub_dt.hour)
        data["monthday"].append(pub_dt.day)
        # data["week"].append(pub_dt.isocalendar().week)
        data["date"].append(pub_dt.date().isoformat())
        data["datemonth"].append(pub_dt.date().isoformat()[:-3])

    data.pop("published")

    logging.info(f"Dates range from {earliest_pub} to {latest_pub}")

    if "stats" in args.print:
        tables = []
        for k, v in data.items():
            t = Table(title=k)
            t.add_column("name", style="green")
            t.add_column("#", style="cyan")
            t.add_column("", style="blue")
            c = Counter(v)
            if k in ["month", "hour", "monthday", "week", "date", "datemonth"]:
                to_be_shown = sorted(c.items())
            elif k == "weekday":
                to_be_shown = [(k, c[k]) for k in weekdays]
            else:
                to_be_shown = c.most_common(30)

            for item in to_be_shown:
                if (
                    k not in ["date", "datemonth"]
                    or (k == "date" and item[0] >= "2021-12")
                    or (k == "datemonth" and item[0] >= "2021")
                ):
                    t.add_row(*map(str, item), bar(item[1] / c.most_common(1)[0][1], 20))

            tables.append(t)

        console.rule("[bold orange1]Statistics[/]", style="white")
        print(Columns(tables))

    if args.print:
        console.rule("[bold orange1]Articles[/]", style="white")
        if "title" in args.print and "text" not in args.print and "fulltext" not in args.print:
            for art in found_articles:
                category = art["category"][0] if art["category"] else None
                print(
                    f'{datetime.fromisoformat(art["published"]).ctime()} | {art["outlet"]} '
                    f'| {category} | {", ".join(art["authors"])}'
                )
                print(art["url"])
                print(f'[bold]{art["title"]}[/]\n')
        if "text" in args.print or "fulltext" in args.print:
            panels = []
            h = None if "fulltext" in args.print else 20
            for art in found_articles:
                text = Text(art["text"], style="default")
                for i, key in enumerate(args.full_text):
                    text.highlight_words([key], colors[i], case_sensitive=False)
                content = Text.assemble(
                    (f'{art["title"]}\n', "bold white"),
                    (f'({art["url"]})\n\n', f'dim link {art["url"]}'),
                    text,
                )
                panels.append(
                    Panel(
                        content,
                        width=100,
                        height=h,
                        padding=1,
                        title=f'{datetime.fromisoformat(art["published"]).ctime()} '
                        f'| {art["outlet"]} | {art["category"][0]} | {", ".join(art["authors"])}',
                    )
                )

            print(Columns(panels))

    if args.save:
        logging.info(f"Saving data {args.save} from {len(found_articles)} articles to {args.output_file}")
        try:
            f = open(args.output_file, "w")
        except Exception as e:
            raise e

        for art in found_articles:
            logging.info(f"{art['published']}\t{art['title']}")
            if "title" in args.save:
                f.write(art["title"] + "\n")
            if "text" in args.save:
                f.write(art["text"] + "\n")
        f.close()


if __name__ == "__main__":
    run()



"""Tools for logging and text output
"""

import functools
import inspect
import logging
from datetime import datetime
from time import perf_counter

from rich.box import DOUBLE_EDGE
from rich.console import Console
from rich.logging import RichHandler
from rich.panel import Panel
from rich.text import Text
from rich.traceback import install

console = Console(record=True)

install(show_locals=True)

# from rich.table import Table
# from rich.highlighter import RegexHighlighter, Highlighter
# from rich.theme import Theme

# class IsoformatDateHighlighter(RegexHighlighter):
#     """Apply style to anything that looks like an email."""

#     base_style = "monito."
#     highlights = [r"?P<isodate>config"]
#     # highlights = [r'^(?P<isodate>-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])-(3[01]|0[1-9]|[12][0-9])'
#           'T(2[0-3]|[01][0-9]):([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$']

# class RainbowHighlighter(Highlighter):
#     def highlight(self, text):
#         for index in range(len(text)):
#             text.stylize(f"color({randint(16, 255)})", index, index + 1)


# rainbow = RainbowHighlighter()


log = logging.getLogger("rich")

logging.getLogger("selenium").setLevel(logging.WARNING)
logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)

_timers = {}


def logging_set_up(log_to_stdout: bool = True):
    from monito_tools import general_config

    if general_config.get("log_to_file_width", 0) > 0:
        console.width = general_config["log_to_file_width"]
    if log_to_stdout:
        general_config["log_to_file"] = ""
        general_config["log_filename"] = "log"
    if general_config["log_to_file"]:
        general_config["log_filename"] = general_config["log_to_file"].replace(
            "$DATE", datetime.today().strftime("%Y-%m-%d_%H-%M-%S")
        )
        try:
            console.file = open(general_config["log_filename"], "w")
        except Exception as e:
            log.error(f"Error opening {general_config['log_filename']} for writing. Falling back to stdout")
            log.error(e)
        else:
            general_config["hide_progress"] = True

    logging.basicConfig(
        level="INFO",
        format="%(message)s",
        datefmt="[%F %T]",
        # handlers=[RichHandler(console=Console(highlighter=IsoformatDateHighlighter(),
        #                                       theme=Theme({"monito.isodate": "bold magenta"})),
        #                       rich_tracebacks=True)],
        handlers=[RichHandler(rich_tracebacks=True, console=console, show_path=False, show_time=False)],
    )


def set_log_level(quiet: bool = False, verbose: bool = False, debug: bool = False) -> None:  # pragma: no cover
    if debug is True:
        log.setLevel(logging.DEBUG)
        log.info("debug mode engaged")
    if verbose is True:
        log.setLevel(logging.INFO)
    if quiet is True:
        log.setLevel(logging.ERROR)


def print_logo(version: str):
    """print the logo and version"""
    console.print(
        Panel(
            Text(f"monito {version}", justify="center", style="bold red"),
            box=DOUBLE_EDGE,
            highlight=True,
        )
    )


def profile(func):
    """profiling and logging decorator"""
    stk = inspect.stack()[1]
    mod = inspect.getmodulename(stk.filename)

    # print(mod)
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        t0 = perf_counter()
        log.debug(f"Entering {mod}.{func.__name__}")
        ret = func(*args, **kwargs)
        log.info(f"Exiting {mod}.{func.__name__}; {perf_counter() - t0:.2f} s")
        return ret

    return wrapper


def debug(func):
    """Print the function signature and return value"""

    @functools.wraps(func)
    def wrapper_debug(*args, **kwargs):
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        log.debug(f"Calling {func.__name__}({signature})")
        value = func(*args, **kwargs)
        log.debug(f"{func.__name__!r} returned {value!r}")
        return value

    return wrapper_debug


def timer_start(name: str = "default"):
    """start a named timer"""
    _timers[name] = perf_counter()


def timer_stop(name: str = "default"):
    """print the time of a named timer"""
    if name not in _timers:
        log.warning(f"timer {name} has not bee started")
    log.info(f"timer '{name}': {perf_counter() - _timers[name]:.3f} s")



"""Find articles similar to a given text file
"""
import pickle
from datetime import datetime

from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity

from monito_tools import init_mongodb, load_general_config

# client = MongoClient("mongodb://localhost:27017")
# database = client["monite"]
# articles_db = database["articles"]

# load the general configuration
log_msgs = load_general_config("monito_config.toml")

# init data base
articles_db = init_mongodb()


start = "2023-04-01"
end = "2023-04-08"
vectorizer_file = "data/vectorizer_2023-01-06_35087.pk"
vectorizer = None
sensitivity = 0.99

create_vectorizer = False
compare = True

if create_vectorizer:
    print(articles_db)
    cursor = articles_db.find({"published": {"$gte": "2023-03", "$lte": "2023-04"}})
    corpus = [c["title"] + "\n" + c["text"] for c in cursor]

    # print(corpus[:2])

    vectorizer = TfidfVectorizer()  # TODO: add options stop_words, ngram_range, max_features
    vectorizer.fit(corpus)
    filename = f"vectorizer_{datetime.today().strftime('%Y-%m-%d')}_{len(corpus)}.pk"
    with open(filename, "wb") as f:
        pickle.dump(vectorizer, f)

        print(f"{datetime.today().strftime('[%Y-%m-%d %H:%M:%S]')} saved the vectorizer as {filename}")
        vectorizer_file = filename
    print(f"Saved vectorizer as {filename}.")
    exit()

if compare:
    # print("article:")
    # article = sys.stdin.read()

    published = []
    outlets = []
    titles = []
    texts = []
    urls = []

    cursor = articles_db.find({"published": {"$gte": start, "$lte": end}})

    for i, c in enumerate(cursor):
        # print(i, "\t", c["published"], c["title"])

        published.append(c["published"])
        outlets.append(c["outlet"])
        titles.append(c["title"])
        texts.append(c["text"])
        urls.append(c["url"])

    if vectorizer is None:
        with open(vectorizer_file, "rb") as f:
            vectorizer = pickle.load(f)

    matrix_text = vectorizer.transform(texts)
    sim_matrix = cosine_similarity(matrix_text, matrix_text)

    print(sim_matrix)

    count = len(texts)
    count_groups = 0
    already_processed = set()
    for i in range(count):
        printed_first = False
        if i in already_processed:
            continue
        for j in range(count):
            if i == j:
                continue
            if j in already_processed:
                continue
            if sim_matrix[i][j] > sensitivity:
                if not printed_first:
                    print(f"\n{i}\t{outlets[i]:30} {titles[i]} {urls[i]}")
                    count_groups += 1
                    printed_first = True
                # print(round(sim_matrix[i][j], 3), i, j)
                print(f"{j}\t{outlets[j]:30} {titles[j]} {urls[j]}")
                already_processed.add(j)
        already_processed.add(i)

    print("\n", count_groups)
